//******************************************************************
//* SILICON LABS CP210x DEMO
//* WRITTEN BY FRED EADY FOR NUTS AND VOLTS MAGAZINE
//* LAST UPDATED 01/06/2009
//* CHANGES/ADDITIONS
//******************************************************************b
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>

// Settings for PIC18F4620
__CONFIG(1,IESODIS & FCMDIS & HS);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0106);
//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void putch(char c);
char recvchar(void);
char sendchar(char data);
char CharInQueue(void);
//******************************************************************
//*	PIC18F4620 RAM Definitions
//******************************************************************
//******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
#define BAUD	57600		//desired baud rate
#define FOSC	20000000	//oscillator frequency
#define DIVIDER ((unsigned int)(FOSC/(16 * BAUD) - 1))
//1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_RX_BUFFER_SIZE  256     
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_TX_BUFFER_SIZE  256      
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
char USART_RxBuf[USART_RX_BUFFER_SIZE],USART_TxBuf[USART_TX_BUFFER_SIZE];
char USART_TxHead,USART_TxTail,USART_RxHead,USART_RxTail;
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}

//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt TIMERS(void)
{
  char usart_data,tmphead,tmptail;

  if(RCIF)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
   USART_RxHead = tmphead;        // store new index 

   if ( tmphead == USART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   USART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT)
  {
                                // check if all usart_data is transmitted  
   if ( USART_TxHead != USART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
     USART_TxTail = tmptail;      // store new index  

    TXREG = USART_TxBuf[tmptail];  // start transmission
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }
} 
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
// load baud rate divisor for 576000 bps with 20MHz clock  
  SPBRG = DIVIDER;	
  TRISC7 = 1;		//receive pin
  TRISC6 = 0;		//transmit pin
  TXSTA = 0x04;		//high speed baud rate	
  RCSTA = 0x80;		//enable serial port and serial port pins
  
  USART_RxTail = 0x00;	//flush receive buffer 
  USART_RxHead = 0x00;
  USART_TxTail = 0x00;	//flush transmit buffer
  USART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable USART1 receiver
  TXIE = 0;			//disable USART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init(void)
{
   	TRISA = 0b11111111;		
   	TRISB = 0b11111111;		
   	TRISC = 0b11111111;
//******************************************************************
//*	CONFIGURE EEPROM READ/WRITE
//******************************************************************
	EECON1 = 0b00000000;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON0 = 0b00000000;	//turn OFF A2D
	ADCON1 = 0b00001111;	//all digital
//******************************************************************
//*	INITIALIZE THE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	ACTIVATE INTERRUPTS
//******************************************************************
    enable_GLOBALint;
//******************************************************************
}
//******************************************************************
//*	GET A CHARACTER FROM RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
                                // wait for incomming usart_data  
  while ( USART_RxHead == USART_RxTail );
                                // calculate buffer index  
  tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxTail = tmptail;        // store new index  

  return USART_RxBuf[tmptail];   // return usart_data  
}

//******************************************************************
//*	SEND A CHARACTER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == USART_TxTail );
                                // store usart_data in buffer 
  USART_TxBuf[tmphead] = usart_data;
  USART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(USART_RxHead != USART_RxTail);
}
//******************************************************************
//*	MAIN FUNCTION
//******************************************************************
void main(void)
{
	unsigned int x,i;
	
  	init();				//init the EUSART
  	printf("NUTS AND VOLTS USB ECHO APPLICATION\r\n");
  	for(i=0;i<0x3FFF;++i)
  	{
	  	++x;
	  }
  	do{
	  	while(!(CharInQueue()));
	  	sendchar(recvchar());
	  }while(1);	
}
